#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>

void services(void);
void led_drive(void);
void key_scan(void);
void init(void);
void cls(void);
void putnum(char c);
void allclear(void);
void clear(void);
void calc_plus(void);
void calc_minus(void);
void calc_mul(void);
void calc_div(void);
void execcalc(void);
void putdot(void);
void changesign(void);
void squareroot(void);
void disp(void);
double sqrt_nr(double a);
double my_atof(char *s);

/*
    LED I/O蓖                       KEY I/O蓖

    PC:     0 ... digit 0 enable            PD:     0 ... key column 0 strobe
            :           :                           :                :
            7 ... digit 7 enable                    4 ... key column 4 strobe

    PA:       -7-                           PB:     0 ... key row 0 read
            3|   |6                                 :             :
              -2-                                   3 ... key row 3 read
            1|   |5
              -0- .4
*/

char const led_pattern[19] = {
	0b00010100,  /*  0  */
	0b10011111,  /*  1  */
	0b00111000,  /*  2  */
	0b00011010,  /*  3  */
	0b10010011,  /*  4  */
	0b01010010,  /*  5  */
	0b01010000,  /*  6  */
	0b00011111,  /*  7  */
	0b00010000,  /*  8  */
	0b00010010,  /*  9  */
	0b00010001,  /*  A  */
	0b11010000,  /*  b  */
	0b11111000,  /*  c  */
	0b10011000,  /*  d  */
	0b01110000,  /*  E  */
	0b01110001,  /*  F  */
	0b11111011,  /*  -  */
	0b11101111,  /*  .  */
	0b11111111,  /*     */
};

char const key_table[4][5] = {
	{ '7', '8', '9', '/', 'C',  },
	{ '4', '5', '6', '*', 'e',  },
	{ '1', '2', '3', '-', 'r',  },
	{ '0', '.', 's', '+', '=',  },
};

char disp_data[8];
int column;
int strobe;
int inkey;
int keystat;
int prev_keystat;
char num_a[10],num_b[10],buf[10],op;
int ent;
int errflag;

void services(void)	/* GpTu[` */
{
	led_drive();	/* LED쓮 */
	key_scan();		/* L[ XL */
}

void led_drive(void)	/* LED쓮 */
{
	PORTA = 0xFF;
	PORTC = 1<<column;
	PORTA = led_pattern[ (int)disp_data[column] ];
	column = (column+1) % 8;
}

void key_scan(void)		/* L[ XL */
{
	int i,k;

	PORTD = ~( 1<< strobe );	/* Xg[uMo */
	for (i=0;i<100;i++) {	/* ҂ */
		;
	}
	k = PINB;	/* L[ f[^ ǂݏo */
	k = (k ^ 0xFF) & 0x0F;
	for (i=0;i<4;i++) {
		if (k & (1<<i)) { keystat = key_table[i][strobe]; }
	}
	strobe = (strobe+1) % 5;
	if (strobe==0) {
		if ((keystat == 0)&&(prev_keystat != 0)) {
			inkey = prev_keystat;
		} else {
			inkey = 0;
		}
		prev_keystat = keystat;
		keystat=0;
	}
}

void io_init(void)	/* I/O|[g̏ */
{
	DDRA |= 0b11111111; /* In/Out direction  1...Output / 0...Input */
	DDRB |= 0b00000000; /* In/Out direction  1...Output / 0...Input */
	DDRC |= 0b11111111; /* In/Out direction  1...Output / 0...Input */
	DDRD |= 0b11111111; /* In/Out direction  1...Output / 0...Input */

	PORTA = 0xFF; /* LED 8seg */
	PORTC = 0xFF; /* LED -ʒu */
	PORTB = 0xFF; /* KEY read */
	PORTD = 0xFF; /* KEY strobe */
}

void init(void)		/*  */
{
	/* e평Tu[`Ăяo */
	io_init();
	/* ϐ̏ */
	column=0;
	strobe=0;
	keystat=0;
}

/*****************************
    d̃CvO
*****************************/
int main()
{
	init();
	allclear();
	while (1) {
		services();
		if (inkey != 0) {
			if ((inkey>='0')&&(inkey<='9')) { putnum(inkey); }
			if (inkey=='.') { putdot();     /* _ */ }
			if (inkey=='C') { allclear();   /* I[NA */ }
			if (inkey=='e') { clear();      /* NA */ }
			if (inkey=='+') { calc_plus();  /* Z */ }
			if (inkey=='-') { calc_minus(); /* Z */ }
			if (inkey=='*') { calc_mul();   /* |Z */ }
			if (inkey=='/') { calc_div();   /* Z */ }
			if (inkey=='=') { execcalc();   /* Zs */ }
			if (inkey=='s') { changesign(); /* ](PZ) */ }
			if (inkey=='r') { squareroot(); /* (PZ) */}
			inkey=0;
		}
	}
	return 0;
}

void putnum(char c)
{
	if (ent==0) {
		strcpy(num_b,num_a);
		strcpy(num_a,"0");
	}
	if (strcmp(num_a,"0")==0) {
		num_a[0]=c;
	} else {
		if (strlen(num_a)<8) {
			sprintf(buf,"%s%c",num_a,c);
			strcpy(num_a,buf);
		}
	}
	ent=1;
	disp();
}

void putdot(void)
{
	if (ent==0) {
		strcpy(num_b,num_a);
		strcpy(num_a,"0");
	}
	if (strchr(num_a,'.')==NULL) {
		strcat(num_a,".");
	}
	ent=1;
	disp();
}


void allclear(void)
{
	strcpy(num_a,"0");
	strcpy(num_b,"0");
	op=0;
	ent=0;
	errflag=0;
	disp();
}

void clear(void)
{
	if (errflag==1) {
		errflag=0;
	} else {
		strcpy(num_a,"0");
	}
	disp();
}

void calc_plus(void)
{
	if ((op!=0)&&(ent!=0)) { execcalc(); }
	op='+';
	ent=0;
}

void calc_minus(void)
{
	if ((op!=0)&&(ent!=0)) { execcalc(); }
	op='-';
	ent=0;
}

void calc_mul(void)
{
	if ((op!=0)&&(ent!=0)) { execcalc(); }
	op='*';
	ent=0;
}

void calc_div(void)
{
	if ((op!=0)&&(ent!=0)) { execcalc(); }
	op='/';
	ent=0;
}

void execcalc(void)
{
	/* num_a  <== num_b op num_a */
	double a,b;
	errflag=0;
	a=my_atof(num_a); /* sscanf(num_a,"%lf",&a); */
	b=my_atof(num_b); /* sscanf(num_b,"%lf",&b); */
	if (op=='+') { a = b + a; }
	if (op=='-') { a = b - a; }
	if (op=='*') { a = b * a; }
	if (op=='/') {
		if (a==0){ a = 0; errflag=1; /* 0Z! */ }
		else     { a = b / a; }
	}
	strcpy(num_b,"0");
	sprintf(num_a,"%g",a);
	op=0;
	disp();
}

void changesign(void)
{
	/* num_a  <==  (-1) * num_a  */
	double a;
	a=my_atof(num_a); /* sscanf(num_a,"%lf",&a); */
	a = -a;
	sprintf(num_a,"%g",a);
	ent=0;
	disp();
}

void squareroot(void)
{
	/* num_a  <==  sqrt( num_a )  */
	double a;
	a=my_atof(num_a); /* sscanf(num_a,"%lf",&a); */
	if (a<0) {
		a = -a; /* ̍͋Ȃ̂ŌvZs\BO */
		errflag=1;
	}
	a = sqrt_nr(a);
	sprintf(num_a,"%g",a);
	ent=0;
	disp();
}

double sqrt_nr(double a)
{
	/* Newton-Raphson@ŌvZ */
	double x,x0,d;
	if (a==0) { return 0; }
	x0=0;
	x=1;
	while (1) {
		x = x - ( x * x - a ) / ( x * 2 );
		d = x0 - x; if ( d < 0 ) { d = -d; }
		if (d<0.000001) { break; }
		x0 = x;
	}
	return x;
}

double my_atof(char *s)
{
	double result,sign,d;
	result=0; sign=1; d=0.1;
	if (*s=='-')  { sign= -1; s++; }
	while (1) {
		if (*s=='\0') { break; }
		if (*s=='.')  { s++; break; }
		if ((*s>='0')&&(*s<='9')) { result = result*10 + (*s-'0'); }
		s++;
	}
	while (1) {
		if (*s=='\0') { break; }
		if ((*s=='e')||(*s=='E')) { break; } /* 1.0e+3̌`ɂ͖Ή */
		if ((*s>='0')&&(*s<='9')) { result = result + (*s-'0')*d; d/=10; }
		s++;
	}
	result *= sign;
	return result;
}

void disp(void)
{
	/* \pobt@disp_data[]ɕ\eZbg */
	char c,*p;
	int i;
	if ((p=strchr(num_a,'e'))!=NULL) {
		/*  1.23456e+07̂悤Ȍ`ɂȂG[, w𖳎 */
		*p='\0'; errflag=1;
	}
	if (strlen(num_a)>8) {
		num_a[8]='\0';
		if ((strchr(num_a,'.'))==NULL) {
			num_a[7]=num_a[6];		/* Ōӂꂵ珬ɕϊ */
			num_a[6]=num_a[5];
			num_a[5]=num_a[4];
			num_a[4]=num_a[3];
			num_a[3]=num_a[2];
			num_a[2]=num_a[1];
			num_a[1]='.';
			errflag=1;
		}
	}
	for (i=0;i<8;i++) {
		disp_data[i]=0x12;		/* ܂ŠNA */
	}
	for (i=0;i<strlen(num_a);i++) {
		c = num_a[i];
		if ((c>='0')&&(c<='9')) { c = c-'0'; }
		if ((c>='A')&&(c<='F')) { c = c-'A'+10; }
		if ((c>='a')&&(c<='f')) { c = c-'a'+10; }
		if (c=='-') { c=0x10; }
		if (c=='.') { c=0x11; }
		disp_data[7-i] = c;		/* l߂ŕ\ */
	}
	if (errflag==1) {			/* G[ɂ͈ԉĚE\ */
		disp_data[0]=0x0E;
	}
}
